class RouteWithTooSmallCapacity(Exception):
    pass


class RebalanceFailure(Exception):
    pass


class NoRouteError(Exception):
    pass


class DryRunException(Exception):
    pass


class PaymentTimeOut(Exception):
    pass


class TooExpensive(Exception):
    pass


#!/usr/bin/env python
import argparse
import time
import os

from lndmanage.lib.node import LndNode
from lndmanage.lib.listchannels import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.recommend_nodes import RecommendNodes
from lndmanage.lib.exceptions import (
    DryRunException,
    PaymentTimeOut,
    TooExpensive,
    RebalanceFailure
)
from lndmanage import settings

import logging.config
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()


def range_limited_float_type(unchecked_value):
    """
    Type function for argparse - a float within some predefined bounds

    :param: unchecked_value: float
    """
    try:
        value = float(unchecked_value)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if value < 1E-6 or value > 1:
        raise argparse.ArgumentTypeError(
            "Argument must be < " + str(1E-6) + " and > " + str(1))
    return value


def unbalanced_float(x):
    """
    Checks if the value is a valid unbalancedness between [-1 ... 1]
    """
    x = float(x)
    if x < -1.0 or x > 1.0:
        raise argparse.ArgumentTypeError(f"{x} not in range [-1.0, 1.0]")
    return x


class Parser(object):
    def __init__(self):
        # setup the command line parser
        self.parser = argparse.ArgumentParser(
            prog='lndmanage.py',
            description='Lightning network daemon channel management tool.')
        self.parser.add_argument(
            '--loglevel', default='INFO', choices=['INFO', 'DEBUG'])
        subparsers = self.parser.add_subparsers(dest='cmd')

        # cmd: status
        self.parser_status = subparsers.add_parser(
            'status', help='display node status',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)

        # cmd: listchannels
        self.parser_listchannels = subparsers.add_parser(
            'listchannels',
            help='lists channels with extended information '
                 '[see also subcommands with -h]',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        listchannels_subparsers = self.parser_listchannels.add_subparsers(
            dest='subcmd')

        # subcmd: listchannels rebalance
        parser_listchannels_rebalance = listchannels_subparsers.add_parser(
            'rebalance', help='displays unbalanced channels')
        parser_listchannels_rebalance.add_argument(
            '--unbalancedness', type=float,
            default=settings.UNBALANCED_CHANNEL,
            help='Unbalancedness is a way to express how balanced a '
                 'channel is, a value between [-1, 1] (a perfectly balanced '
                 'channel has a value of 0). The flag excludes channels with '
                 'an absolute unbalancedness smaller than UNBALANCEDNESS.')
        parser_listchannels_rebalance.add_argument(
            '--sort-by', default='rev_ub', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels inactive
        parser_listchannels_inactive = listchannels_subparsers.add_parser(
            'inactive', help="displays inactive channels")
        parser_listchannels_inactive.add_argument(
            '--sort-by', default='lup', type=str,
            help='sort by column (look at description)')

        # subcmd: listchannels forwardings
        parser_listchannels_forwardings = listchannels_subparsers.add_parser(
            'forwardings',
            help="displays channels with forwarding information")
        parser_listchannels_forwardings.add_argument(
            '--from-days-ago', default=365, type=int,
            help='time interval start (days ago)')
        parser_listchannels_forwardings.add_argument(
            '--to-days-ago', default=0, type=int,
            help='time interval end (days ago)')
        parser_listchannels_forwardings.add_argument(
            '--sort-by', default='f/w', type=str,
            help='sort by column (look at description)')

        # cmd: rebalance
        self.parser_rebalance = subparsers.add_parser(
            'rebalance', help='rebalance a channel',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_rebalance.add_argument('channel', type=int,
                                           help='channel_id')
        self.parser_rebalance.add_argument(
            '--max-fee-sat', type=int, default=20,
            help='Sets the maximal fees in satoshis to be paid.')
        self.parser_rebalance.add_argument(
            '--chunksize', type=float, default=1.0,
            help='Specifies if the individual rebalance attempts should be '
                 'split into smaller relative amounts. This increases success'
                 ' rates, but also increases costs!')
        self.parser_rebalance.add_argument(
            '--max-fee-rate', type=range_limited_float_type, default=5E-5,
            help='Sets the maximal effective fee rate to be paid.'
                 ' The effective fee rate is defined by '
                 '(base_fee + amt * fee_rate) / amt.')
        self.parser_rebalance.add_argument(
            '--reckless', help='Execute action in the network.',
            action='store_true')
        self.parser_rebalance.add_argument(
            '--allow-unbalancing',
            help=f'Allow channels to get an unbalancedness'
            f' up to +-{settings.UNBALANCED_CHANNEL}.',
            action='store_true')
        self.parser_rebalance.add_argument(
            '--target', help=f'This feature is still experimental! '
            f'The unbalancedness target is between [-1, 1]. '
            f'A target of -1 leads to a maximal local balance, a target of 0 '
            f'to a 50:50 balanced channel and a target of 1 to a maximal '
            f'remote balance. Default is a target of 0.',
            type=unbalanced_float, default=None)
        rebalancing_strategies = ['most-affordable-first',
                                  'lowest-feerate-first', 'match-unbalanced']
        self.parser_rebalance.add_argument(
            '--strategy',
            help=f'Rebalancing strategy.',
            choices=rebalancing_strategies, type=str, default=None)

        # cmd: circle
        self.parser_circle = subparsers.add_parser(
            'circle', help='circular self-payment',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_circle.add_argument('channel_from', type=int,
                                        help='channel_from')
        self.parser_circle.add_argument('channel_to', type=int,
                                        help='channel_from')
        self.parser_circle.add_argument('amt_sat', type=int,
                                        help='amount in satoshis')
        self.parser_circle.add_argument(
            '--max-fee-sat', type=int, default=20,
            help='Sets the maximal fees in satoshis to be paid.')
        self.parser_circle.add_argument(
            '--max-fee-rate', type=range_limited_float_type, default=5E-5,
            help='Sets the maximal effective fee rate to be paid. '
                 'The effective fee rate is defined by '
                 '(base_fee + amt * fee_rate) / amt.')
        self.parser_circle.add_argument(
            '--reckless', help='Execute action in the network.',
            action='store_true')

        # cmd: recommend-nodes
        self.parser_recommend_nodes = subparsers.add_parser(
            'recommend-nodes',
            help='recommends nodes [see also subcommands with -h]',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        self.parser_recommend_nodes.add_argument(
            '--show-connected', action='store_true', default=False,
            help='specifies if already connected nodes should be '
                 'removed from list')
        self.parser_recommend_nodes.add_argument(
            '--show-addresses', action='store_true', default=False,
            help='specifies if node addresses should be shown')
        parser_recommend_nodes_subparsers = \
            self.parser_recommend_nodes.add_subparsers(
                dest='subcmd')

        # TODO: put global options to the
        #  parent parser (e.g. number of nodes, sort-by flag)

        # subcmd: recommend-nodes good-old
        parser_recommend_nodes_good_old = \
            parser_recommend_nodes_subparsers.add_parser(
                'good-old',
                help='shows nodes already interacted with but no '
                     'active channels',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_good_old.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_good_old.add_argument(
            '--sort-by', default='tot', type=str,
            help="sort by column [abbreviation, e.g. 'tot']")

        # subcmd: recommend-nodes flow-analysis
        parser_recommend_nodes_flow_analysis = \
            parser_recommend_nodes_subparsers.add_parser(
                'flow-analysis', help='recommends nodes from a flow analysis',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_flow_analysis.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--forwarding-events', default=200, type=int,
            help='sets the number of forwarding events in the flow analysis')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--inwards', action='store_true',
            help='if True, inward-flowing nodes are displayed '
                 'instead of outward-flowing nodes')
        parser_recommend_nodes_flow_analysis.add_argument(
            '--sort-by', default='weight', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes external_source
        parser_recommend_nodes_external_source = \
            parser_recommend_nodes_subparsers.add_parser(
                'external-source', help='recommends nodes from a given file/url',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_external_source.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_external_source.add_argument(
            '--source', type=str,
            default='https://github.com/lightningnetworkstores/'
                    'lightningnetworkstores.github.io/raw/master/sites.json',
            help='url/file to be analyzed')
        parser_recommend_nodes_external_source.add_argument(
            '--distributing-nodes', action='store_true',
            help='if True, distributing nodes are '
                 'displayed instead of the bare nodes')
        parser_recommend_nodes_external_source.add_argument(
            '--sort-by', default='cpc', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

        # subcmd: recommend-nodes channel-openings
        parser_recommend_nodes_channel_openings = \
            parser_recommend_nodes_subparsers.add_parser(
                'channel-openings',
                help='recommends nodes from recent channel openings',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        parser_recommend_nodes_channel_openings.add_argument(
            '--nnodes', default=20, type=int,
            help='sets the number of nodes displayed')
        parser_recommend_nodes_channel_openings.add_argument(
            '--from-days-ago', type=int,
            default=30,
            help='channel openings starting from a time frame days ago')
        parser_recommend_nodes_channel_openings.add_argument(
            '--sort-by', default='msteady', type=str,
            help="sort by column [abbreviation, e.g. 'nchan']")

    def parse_arguments(self):
        return self.parser.parse_args()


def main():
    parser = Parser()
    args = parser.parse_arguments()
    # print(args)

    if args.cmd is None:
        parser.parser.print_help()
        return 0

    # program execution
    if args.loglevel:
        # update the loglevel of the stdout handler to the user choice
        logger.handlers[0].setLevel(args.loglevel)

    # config.ini is expected to be in home/.lndmanage directory
    config_file = os.path.join(settings.home_dir, 'config.ini')
    node = LndNode(config_file=config_file)

    if args.cmd == 'status':
        node.print_status()

    elif args.cmd == 'listchannels':
        listchannels = ListChannels(node)
        if not args.subcmd:
            listchannels.print_all_channels('rev_alias')
        if args.subcmd == 'rebalance':
            listchannels.print_channels_unbalanced(
                args.unbalancedness, sort_string=args.sort_by)
        elif args.subcmd == 'inactive':
            listchannels.print_channels_inactive(
                sort_string=args.sort_by)
        elif args.subcmd == 'forwardings':
            # convert time interval into unix timestamp
            time_from = time.time() - args.from_days_ago * 24 * 60 * 60
            time_to = time.time() - args.to_days_ago * 24 * 60 * 60
            listchannels.print_channels_forwardings(
                time_interval_start=time_from, time_interval_end=time_to,
                sort_string=args.sort_by)

    elif args.cmd == 'rebalance':
        if args.target:
            logger.warning("Warning: Target is set, this is still an "
                           "experimental feature.")
        rebalancer = Rebalancer(node, args.max_fee_rate, args.max_fee_sat)
        try:
            rebalancer.rebalance(
                args.channel, dry=not args.reckless, chunksize=args.chunksize,
                target=args.target, allow_unbalancing=args.allow_unbalancing,
                strategy=args.strategy)
        except RebalanceFailure as e:
            logger.error(f"Error: {e}")

    elif args.cmd == 'circle':
        rebalancer = Rebalancer(node, args.max_fee_rate, args.max_fee_sat)
        invoice_r_hash = node.get_rebalance_invoice(memo='circular payment')
        try:
            rebalancer.rebalance_two_channels(
                args.channel_from, args.channel_to,
                args.amt_sat, invoice_r_hash, args.max_fee_sat,
                dry=not args.reckless)
        except DryRunException:
            logger.info("This was just a dry run.")
        except TooExpensive:
            logger.error("Payment failed. This is likely due to a too low "
                         "default --max-fee-rate.")
        except PaymentTimeOut:
            logger.error("Payment failed because the payment timed out. "
                         "This is an unresolved issue.")

    elif args.cmd == 'recommend-nodes':
        if not args.subcmd:
            parser.parser_recommend_nodes.print_help()
            return 0

        recommend_nodes = RecommendNodes(
            node, show_connected=args.show_connected,
            show_addresses=args.show_addresses)

        if args.subcmd == 'good-old':
            recommend_nodes.print_good_old(number_of_nodes=args.nnodes,
                                           sort_by=args.sort_by)
        elif args.subcmd == 'flow-analysis':
            recommend_nodes.print_flow_analysis(
                out_direction=(not args.inwards),
                number_of_nodes=args.nnodes,
                forwarding_events=args.forwarding_events,
                sort_by=args.sort_by)
        elif args.subcmd == 'external-source':
            recommend_nodes.print_external_source(
                args.source, distributing_nodes=args.distributing_nodes,
                number_of_nodes=args.nnodes, sort_by=args.sort_by)
        elif args.subcmd == 'channel-openings':
            recommend_nodes.print_channel_openings(
                from_days_ago=args.from_days_ago,
                number_of_nodes=args.nnodes, sort_by=args.sort_by)


if __name__ == '__main__':
    main()



"""
Tests for circular self-payments.
"""
import time
from unittest import TestCase

from lnregtest.lib.network import RegtestNetwork

from lndmanage.lib.node import LndNode
from lndmanage.lib.listchannels import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.exceptions import RebalanceFailure, TooExpensive
from lndmanage import settings

from test.testing_common import (
    bin_dir,
    test_data_dir,
    test_graphs_paths,
    lndmanage_home,
    SLEEP_SEC_AFTER_REBALANCING
)

import logging.config
settings.set_lndmanage_home_dir(lndmanage_home)
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.handlers[0].setLevel(logging.DEBUG)


class TestCircle(TestCase):
    def setUp(self):
        self.testnet = RegtestNetwork(
            binary_folder=bin_dir,
            network_definition_location=test_graphs_paths['small_star_ring'],
            nodedata_folder=test_data_dir,
            node_limit='H',
            from_scratch=True
        )
        self.testnet.run_nocleanup()

        master_node_data_dir = self.testnet.master_node.lnd_data_dir
        master_node_port = self.testnet.master_node.grpc_port
        master_node_networkinfo = self.testnet.master_node.getnetworkinfo()

        # assert some basic properties of the graph
        self.assertEqual(4, master_node_networkinfo['num_nodes'])
        self.assertEqual(6, master_node_networkinfo['num_channels'])

        self.lndnode = LndNode(
            lnd_home=master_node_data_dir,
            lnd_host='localhost:' + str(master_node_port),
            regtest=True
        )

        self.rebalancer = Rebalancer(
            self.lndnode,
            max_effective_fee_rate=50,
            budget_sat=20
        )

    def tearDown(self):
        self.testnet.cleanup()

    def circle_and_check(self, rebalancer, channel_number_send,
                         channel_number_receive, amount_sat,
                         expected_fees_msat):
        """
        Helper function for testing a circular payment.

        :param rebalancer:
        :type rebalancer: lndmanage.lib.rebalance.Rebalancer

        :param channel_number_send: channel whose local balance is decreased
        :type channel_number_send: int

        :param channel_number_receive: channel whose local balance is increased
        :type channel_number_receive: int

        :param amount_sat: amount in satoshi to rebalance
        :type amount_sat: int

        :param expected_fees_msat: expected fees in millisatoshi for
            the rebalance
        :type expected_fees_msat: int
        """
        # setup
        graph_before = self.testnet.assemble_graph()
        channel_id_send = self.testnet.channel_mapping[
            channel_number_send]['channel_id']
        channel_id_receive = self.testnet.channel_mapping[
            channel_number_receive]['channel_id']
        invoice_r_hash = self.lndnode.get_invoice(amount_sat, '')

        # exercise
        try:
            fees_msat = rebalancer.rebalance_two_channels(
                channel_id_send,
                channel_id_receive,
                amount_sat,
                invoice_r_hash,
                rebalancer.budget_sat
            )
            time.sleep(SLEEP_SEC_AFTER_REBALANCING)
        except Exception as e:
            raise e

        # check
        graph_after = self.testnet.assemble_graph()
        channel_data_send_before = graph_before['A'][channel_number_send]
        channel_data_receive_before = graph_before['A'][channel_number_receive]
        channel_data_send_after = graph_after['A'][channel_number_send]
        channel_data_receive_after = graph_after['A'][channel_number_receive]
        listchannels = ListChannels(self.lndnode)
        listchannels.print_all_channels('rev_alias')

        # test that the fees are correct
        self.assertEqual(fees_msat, expected_fees_msat)
        # test if sending channel's remote balance has increased correctly
        self.assertEqual(
            amount_sat,
            channel_data_send_after['remote_balance'] -
            channel_data_send_before['remote_balance'] -
            int(expected_fees_msat // 1000),
            "Sending local balance is wrong"
        )
        # test if receiving channel's local balance has increased correctly
        self.assertEqual(
            amount_sat,
            channel_data_receive_after['local_balance'] -
            channel_data_receive_before['local_balance'],
            "Receiving local balance is wrong"
        )

    def test_circle_1_2(self):
        """
        Test successful rebalance from channel 1 to channel 2.
        """
        channel_number_from = 1
        channel_number_to = 2
        amount_sat = 10000
        expected_fees_msat = 43

        self.circle_and_check(
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_1_6(self):
        """
        Test successful rebalance from channel 1 to channel 6.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33

        self.circle_and_check(
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat
        )

    def test_circle_6_1_rebalance_failure_no_funds(self):
        """
        Test expected failure for channel 6 to channel 1, where channel 6
        doesn't have funds.
        """
        channel_number_from = 6
        channel_number_to = 1
        amount_sat = 10000
        expected_fees_msat = 33

        self.assertRaises(
            RebalanceFailure,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_budget_too_expensive(self):
        """
        Test expected failure where rebalance uses more than the fee budget.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33
        self.rebalancer.budget_sat = 0

        self.assertRaises(
            TooExpensive,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_max_fee_rate_too_expensive(self):
        """
        Test expected failure where rebalance is more expensive than
        the desired maximal fee rate.
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 10000
        expected_fees_msat = 33
        self.rebalancer.max_effective_fee_rate = 0

        self.assertRaises(
            TooExpensive,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_fail_first_hop_insufficient_funds(self):
        """
        Tests failure when our channel fails because we didn't take into
        account channel reserve. (Failing channel 1: 117:1:0)
        """
        channel_number_from = 1
        channel_number_to = 6
        amount_sat = 1000000
        expected_fees_msat = 33

        self.assertRaises(
            RebalanceFailure,
            self.circle_and_check,
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )

    def test_circle_1_6_channel_reserve(self):
        """
        Test for a maximal amount circular payment.
        """
        channel_number_from = 1
        channel_number_to = 6
        local_balance = 1000000
        # take into account 1% channel reserve
        amount_sat = int(local_balance - 0.01 * local_balance)
        # need to also subtract commitment fees
        amount_sat -= 9050
        # need to also subtract fees, then error message changes
        amount_sat -= 3
        # extra to make it succeed
        amount_sat -= 2150

        expected_fees_msat = 2938

        self.circle_and_check(
            self.rebalancer,
            channel_number_from,
            channel_number_to,
            amount_sat,
            expected_fees_msat,
        )



import os
from lndmanage import settings

settings.CACHING_RETENTION_MINUTES = 0

# constants for testing
SLEEP_SEC_AFTER_REBALANCING = 2

# testing base folder
test_dir = os.path.dirname(os.path.realpath(__file__))
bin_dir = os.path.join(test_dir, 'bin')
graph_definitions_dir = os.path.join(test_dir, 'graph_definitions')
test_data_dir = os.path.join(test_dir, 'test_data')

# set lndmanage_home path to be in the test_data folder and make sure the
# folder exists
lndmanage_home = os.path.join(test_data_dir, 'lndmanage')
os.makedirs(lndmanage_home, exist_ok=True)

test_graphs_paths = {
    'small_star_ring': os.path.join(
    graph_definitions_dir, 'small_star_ring.py'),
}




"""
Tests for rebalancing of channels.
"""
import time
from unittest import TestCase

from lnregtest.lib.network import RegtestNetwork

from lndmanage import settings
from lndmanage.lib.node import LndNode
from lndmanage.lib.listchannels import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.ln_utilities import channel_unbalancedness_and_commit_fee

from test.testing_common import (
    bin_dir,
    test_data_dir,
    lndmanage_home,
    test_graphs_paths,
    SLEEP_SEC_AFTER_REBALANCING)

import logging.config
settings.set_lndmanage_home_dir(lndmanage_home)
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()
logger.setLevel(logging.INFO)
logger.handlers[0].setLevel(logging.DEBUG)


class TestRebalance(TestCase):
    def setUp(self):
        self.testnet = RegtestNetwork(
            binary_folder=bin_dir,
            network_definition_location=test_graphs_paths['small_star_ring'],
            nodedata_folder=test_data_dir,
            node_limit='H',
            from_scratch=True
        )
        self.testnet.run_nocleanup()
        # to run the lightning network in the background and do some testing
        # here, run:
        # $ lnregtest --nodedata_folder /path/to/lndmanage/test/test_data/
        # self.testnet.run_from_background()

        # logger.info("Generated network information:")
        # logger.info(format_dict(self.testnet.node_mapping))
        # logger.info(format_dict(self.testnet.channel_mapping))
        # logger.info(format_dict(self.testnet.assemble_graph()))

        master_node_data_dir = self.testnet.master_node.lnd_data_dir
        master_node_port = self.testnet.master_node.grpc_port
        master_node_networkinfo = self.testnet.master_node.getnetworkinfo()

        self.assertEqual(4, master_node_networkinfo['num_nodes'])
        self.assertEqual(6, master_node_networkinfo['num_channels'])

        self.lndnode = LndNode(
            lnd_home=master_node_data_dir,
            lnd_host='localhost:' + str(master_node_port),
            regtest=True
        )

        # self.lndnode.print_status()
        logger.info('Initializing done.')

        # channels:
        # A -> B (channel #1)
        # A -> C (channel #2)
        # A -> D (channel #6)
        # B -> C (channel #3)
        # B -> D (channel #4)
        # C -> D (channel #5)

    def tearDown(self):
        self.testnet.cleanup()

    def rebalance_and_check(self, test_channel_number, target,
                            allow_unbalancing, places=5, should_fail=False):
        """
        Test function for rebalancing to a specific target and assert after-
        wards that is was reached.

        :param test_channel_number: int
        :param target: float:
            unbalancedness target
        :param allow_unbalancing: bool:
            unbalancing should be allowed
        :param places: int
            number of digits the result should match to the requested
        :param should_fail: bool:
            indicates whether the rebalancing should fail as requested due to
            maybe unbalancing of other channels
        """
        rebalancer = Rebalancer(
            self.lndnode,
            max_effective_fee_rate=50,
            budget_sat=20
        )

        channel_id = self.testnet.channel_mapping[
            test_channel_number]['channel_id']
        rebalancer.rebalance(
            channel_id,
            dry=False,
            chunksize=1.0,
            target=target,
            allow_unbalancing=allow_unbalancing
        )
        time.sleep(SLEEP_SEC_AFTER_REBALANCING)
        graph = self.testnet.assemble_graph()
        channel_data = graph['A'][test_channel_number]
        listchannels = ListChannels(self.lndnode)
        listchannels.print_all_channels('rev_alias')

        channel_unbalancedness, _ = channel_unbalancedness_and_commit_fee(
            channel_data['local_balance'],
            channel_data['capacity'],
            channel_data['commit_fee'],
            channel_data['initiator']
        )

        if not should_fail:
            self.assertAlmostEqual(target, channel_unbalancedness,
                                   places=places)
        else:
            self.assertNotAlmostEqual(target, channel_unbalancedness,
                                      places=places)

    def test_rebalance_channel_6(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, 0.0, False)

    def test_small_positive_target_channel_6(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, 0.2, False)

    def test_large_positive_channel_6(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, 0.8, False)

    def test_small_negative_target_channel_6_fail(self):
        # this test should fail when unbalancing is not allowed, as it would
        # unbalance another channel if the full target would be accounted for
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, -0.2, False,
                                 should_fail=True)

    def test_small_negative_target_channel_6_succeed(self):
        # this test should fail when unbalancing is not allowed, as it would
        # unbalance another channel if the full target would be accounted for
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, -0.2, True)

    def test_rebalance_channel_1(self):
        test_channel_number = 1
        self.rebalance_and_check(test_channel_number, 0.0, False)

    def test_rebalance_channel_2(self):
        test_channel_number = 2
        self.rebalance_and_check(test_channel_number, 0.0, False, places=1)

    def test_shuffle_arround(self):
        """
        Shuffles sat around in channel 6.
        """
        first_target_amount = -0.1
        second_target_amount = 0.1
        test_channel_number = 6

        self.rebalance_and_check(
            test_channel_number, first_target_amount, True)
        self.rebalance_and_check(
            test_channel_number, second_target_amount, True)



from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ChannelRater(object):
    """
    The purpose of this class is to hold information about the balancedness of channels.
    It also defines the cost function of a channel
    with the function :func:`ChannelRater.lightning_channel_weight`.
    """

    def __init__(self):
        self.bad_channels = {}
        self.bad_nodes = []

    def node_to_node_weight(self, u, v, e, amt_msat):
        """
        Is used to assign a weight for a channel. It is calculated from the fee policy of that channel:
        base fee + proportional fee + path penalty + capacity penalty + blacklist penalty

        :param u: source node
        :param v: target node
        :param e: fee policy of the edges
        :param amt_msat: amount in msat
        :return: cost of the channel in msat
        """
        node_penalty = 0
        if u in self.bad_nodes or v in self.bad_nodes:
            node_penalty = settings.PENALTY

        costs = [node_penalty + self.channel_weight(eattr, amt_msat) for eattr in e.values()]
        return min(costs)

    def channel_weight(self, e, amt_msat):
        long_path_penalty = 0
        if settings.PREFER_SHORT_PATHS:
            long_path_penalty = settings.LONG_PATH_PENALTY_MSAT

        cost = (long_path_penalty
                + e.get('fees')['fee_base_msat']
                + amt_msat * e.get('fees')['fee_rate_milli_msat'] // 1000000
                + self.capacity_penalty(amt_msat, e.get('capacity'))
                + self.disabled_penalty(e.get('fees'))
                + self.already_failed_penalty(e.get('channel_id')))
        return cost

    def add_bad_channel(self, channel, source, target):
        """
        Adds a channel to the blacklist dict.

        :param channel: channel_id
        :param source: pubkey
        :param target: pubkey
        """
        self.bad_channels[channel] = {
            'source': source,
            'target': target,
        }
        logger.debug(f"bad channels so far: {self.get_bad_channels()}")

    def add_bad_node(self, node_pub_key):
        """
        Adds a node public key to the blacklist.

        :param node_pub_key: str
        """
        self.bad_nodes.append(node_pub_key)
        logger.debug(f"bad nodes so far: {self.bad_nodes}")

    def get_bad_channels(self):
        return self.bad_channels.keys()

    def already_failed_penalty(self, channel_id):
        """
        Determines if the channel already failed at some point and penalizes it.

        :param channel_id:
        """
        # TODO: consider also direction
        if channel_id in self.get_bad_channels():
            return settings.PENALTY
        else:
            return 0

    @staticmethod
    def capacity_penalty(amt_msat, capacity_sat):
        """
        Gives a penalty to channels which have too low capacity.

        :param amt_msat
        :param capacity_sat in sat
        :return: penalty
        """
        if capacity_sat < 0.50 * amt_msat // 1000 :
            return settings.PENALTY
        else:
            return 0

    @staticmethod
    def disabled_penalty(policy):
        """
        Gives a penalty to channels which are not active.

        :param policy: policy of the channel, contains state
        :return: high penalty
        """
        if policy['disabled']:
            return settings.PENALTY
        else:
            return 0



import queue

import networkx as nx

from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def ksp_discard_high_cost_paths(graph, source, target, num_k, weight):
    """
    Wrapper for calculating k shortest paths given a weight function and discards paths with penalties.

    :param graph: networkx graph
    :param source: pubkey
    :param target: pubkey
    :param num_k: number of paths
    :param weight: weight function, takes u (pubkey from), v (pubkey to), e (edge information) as arguments
    :return: num_k lists of node pub_keys defining the path
    """
    final_routes = []
    routes, route_costs = ksp(graph, source, target, num_k, weight)
    logger.debug("Approximate costs [msat] of routes:")
    for r, rc in zip(routes, route_costs):
        if rc < settings.PENALTY:
            logger.debug(f"  {rc} msat: {r}")
            final_routes.append(r)
    return final_routes


# ksp algorithm is based on https://gist.github.com/ALenfant/5491853
def path_cost(graph, path, weight=None):
    pathcost = 0
    # print(path)
    for i in range(len(path)):
        if i > 0:
            edge = (path[i-1], path[i])
            if callable(weight):
                try:
                    e = graph[path[i-1]][path[i]]
                except KeyError:
                    e = None
                pathcost += weight(path[i-1], path[i], e)
            else:
                if weight != None:
                    pathcost += graph.get_edge_data(*edge)[weight]
                else:
                    pathcost += 1
    #print("pathcost", pathcost)
    #print()
    return pathcost


# ksp algorithm is based on https://gist.github.com/ALenfant/5491853
def ksp(graph, source, target, num_k, weight=None):
    graph_copy = graph.copy()
    # Shortest path from the source to the target
    A = [nx.shortest_path(graph_copy, source, target, weight=weight)]
    A_costs = [path_cost(graph_copy, A[0], weight)]

    # Initialize the heap to store the potential kth shortest path
    B = queue.PriorityQueue()

    for k in range(1, num_k):
        # The spur node ranges from the first node to the next to last node in the shortest path
        try:
            for i in range(len(A[k-1])-1):
                # Spur node is retrieved from the previous k-shortest path, k - 1
                spurNode = A[k-1][i]
                # The sequence of nodes from the source to the spur node of the previous k-shortest path
                rootPath = A[k-1][:i]

                # We store the removed edges
                removed_edges = []

                for path in A:
                    if len(path) - 1 > i and rootPath == path[:i]:
                        # Remove the links that are part of the previous shortest paths which share the same root path
                        edge = (path[i], path[i+1])
                        if not graph_copy.has_edge(*edge):
                            continue
                        removed_edges.append((edge, graph_copy.get_edge_data(*edge)))
                        graph_copy.remove_edge(*edge)

                # Calculate the spur path from the spur node to the sink
                try:
                    spurPath = nx.shortest_path(graph_copy, spurNode, target, weight=weight)

                    # Entire path is made up of the root path and spur path
                    totalPath = rootPath + spurPath
                    totalPathCost = path_cost(graph_copy, totalPath, weight)
                    # Add the potential k-shortest path to the heap
                    B.put((totalPathCost, totalPath))

                except nx.NetworkXNoPath:
                    pass

                #Add back the edges that were removed from the graph
                #for removed_edge in removed_edges:
                #    print(removed_edge)
                #    graph_copy.add_edge(
                #        *removed_edge[0],
                #        **removed_edge[1]
                #    )

            # Sort the potential k-shortest paths by cost
            # B is already sorted
            # Add the lowest cost path becomes the k-shortest path.
            while True:
                try:
                    cost_, path_ = B.get(False)
                    if path_ not in A:
                        A.append(path_)
                        A_costs.append(cost_)
                        break
                except queue.Empty:
                    break
        except IndexError:
            pass

    return A, A_costs



"""
Module for printing lightning channels.
"""

import math
import logging
from collections import OrderedDict

from lndmanage.lib.forwardings import get_forwarding_statistics_channels
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# define symbols for bool to string conversion
POSITIVE_MARKER = u"\u2713"
NEGATIVE_MARKER = u"\u2717"
ALIAS_LENGTH = 25
ANNOTATION_LENGTH = 25


# define printing abbreviations
# convert key can specify a function, which lets one do unit conversions
PRINT_CHANNELS_FORMAT = {
    'act': {
        'dict_key': 'active',
        'description': 'channel is active',
        'width': 3,
        'format': '^3',
        'align': '>',
        'convert': lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    'age': {
        'dict_key': 'age',
        'description': 'channel age [days]',
        'width': 5,
        'format': '5.0f',
        'align': '>',
    },
    'alias': {
        'dict_key': 'alias',
        'description': 'alias',
        'width': ALIAS_LENGTH,
        'format': '.<'+str(ALIAS_LENGTH),
        'align': '^',
        'convert': lambda x: alias_cutoff(x),
    },
    'annotation': {
        'dict_key': 'annotation',
        'description': 'channel annotation',
        'width': ANNOTATION_LENGTH,
        'format': '.<'+str(ANNOTATION_LENGTH),
        'align': '^',
        'convert': lambda x: alias_cutoff(x),
    },
    'atb': {
        'dict_key': 'amount_to_balanced',
        'description': 'amount to be balanced (local side) [sat]',
        'width': 5,
        'format': '5.0f',
        'align': '>',
    },
    'bf': {
        'dict_key': 'peer_base_fee',
        'description': 'peer base fee [msat]',
        'width': 5,
        'format': '5.0f',
        'align': '>',
    },
    'bwd': {
        'dict_key': 'bandwidth_demand',
        'description': 'bandwidth demand: capacity / max(mean_in, mean_out)',
        'width': 5,
        'format': '5.2f',
        'align': '>',
    },
    'cap': {
        'dict_key': 'capacity',
        'description': 'channel capacity [sat]',
        'width': 9,
        'format': '9d',
        'align': '>',
    },
    'cid': {
        'dict_key': 'chan_id',
        'description': 'channel id',
        'width': 18,
        'format': '',
        'align': '^',
    },
    'fees': {
        'dict_key': 'fees_total',
        'description': 'total fees [sat]',
        'width': 7,
        'format': '7.2f',
        'align': '>',
        'convert': lambda x: float(x) / 1000
    },
    'f/w': {
        'dict_key': 'fees_total_per_week',
        'description': 'total fees per week [sat / week]',
        'width': 6,
        'format': '6.2f',
        'align': '>',
        'convert': lambda x: float(x) / 1000
    },
    'flow': {
        'dict_key': 'flow_direction',
        'description': 'flow direction (positive is outwards)',
        'width': 5,
        'format': '5.2f',
        'align': '>',
        # 'convert': lambda x: '>'*int(nan_to_zero(x)*10/3.0) if x > 0 else
        # '<'*(-int(nan_to_zero(x)*10/3.0))
    },
    'fr': {
        'dict_key': 'peer_fee_rate',
        'description': 'peer fee rate',
        'width': 8,
        'format': '1.6f',
        'align': '>',
        'convert': lambda x: x / 1E6
    },
    'lup': {
        'dict_key': 'last_update',
        'description': 'last update time [days ago]',
        'width': 5,
        'format': '5.0f',
        'align': '>',
    },
    'lb': {
        'dict_key': 'local_balance',
        'description': 'local balance [sat]',
        'width': 9,
        'format': '9d',
        'align': '>',
    },
    'nfwd': {
        'dict_key': 'number_forwardings',
        'description': 'number of forwardings',
        'width': 5,
        'format': '5.0f',
        'align': '>',
    },
    'priv': {
        'dict_key': 'private',
        'description': 'channel is private',
        'width': 5,
        'format': '^5',
        'align': '>',
        'convert': lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    'r': {
        'dict_key': 'action_required',
        'description': 'action is required',
        'width': 1,
        'format': '^1',
        'align': '>',
        'convert': lambda x: NEGATIVE_MARKER if x else '',
    },
    'rb': {
        'dict_key': 'remote_balance',
        'description': 'remote balance [sat]',
        'width': 9,
        'format': '9d',
        'align': '>',
    },
    'in': {
        'dict_key': 'total_forwarding_in',
        'description': 'total forwarding inwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'ini': {
        'dict_key': 'initiator',
        'description': 'true if we opened channel',
        'width': 3,
        'format': '^3',
        'align': '>',
        'convert': lambda x: POSITIVE_MARKER if x else NEGATIVE_MARKER,
    },
    'tot': {
        'dict_key': 'total_forwarding',
        'description': 'total forwarding [sat]',
        'width': 5,
        'format': '5.0f',
        'align': '>',
    },
    'ub': {
        'dict_key': 'unbalancedness',
        'description': 'unbalancedness [-1 ... 1] (0 is 50:50 balanced)',
        'width': 5,
        'format': '5.2f',
        'align': '>',
    },
    'imed': {
        'dict_key': 'median_forwarding_in',
        'description': 'median forwarding inwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'imean': {
        'dict_key': 'mean_forwarding_in',
        'description': 'mean forwarding inwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'imax': {
        'dict_key': 'largest_forwarding_amount_in',
        'description': 'largest forwarding inwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'omed': {
        'dict_key': 'median_forwarding_out',
        'description': 'median forwarding outwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'omean': {
        'dict_key': 'mean_forwarding_out',
        'description': 'mean forwarding outwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'omax': {
        'dict_key': 'largest_forwarding_amount_out',
        'description': 'largest forwarding outwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'out': {
        'dict_key': 'total_forwarding_out',
        'description': 'total forwarding outwards [sat]',
        'width': 10,
        'format': '10.0f',
        'align': '>',
    },
    'sr/w': {
        'dict_key': 'sent_received_per_week',
        'description': 'sent and received per week [sat]',
        'width': 9,
        'format': '9d',
        'align': '>',
    },
}


def alias_cutoff(alias):
    """
    Cuts off the node alias at a certain length and removes unicode
    characters from the node alias.
    :param alias: str
    :return: str
    """
    if len(alias) > ALIAS_LENGTH:
        return alias[:ALIAS_LENGTH-3] + '...'
    else:
        return alias


class ListChannels(object):
    """
    A class to list lightning channels.
    """
    def __init__(self, node):
        """
        :param node: :class:`lib.node.Node`
        """
        self.node = node

    def print_all_channels(self, sort_string='rev_alias'):
        """
        Prints all active and inactive channels.

        :param sort_string: str
        """

        channels = self._add_channel_annotations(self.node.get_all_channels())

        sort_string, reverse_sorting = self._sorting_order(sort_string)
        sort_dict = {
            'function': lambda x: (
                x[1][PRINT_CHANNELS_FORMAT['priv']['dict_key']],
                x[1][sort_string]
                ),
            'string': sort_string,
            'reverse': reverse_sorting,
        }

        self._print_channels(
            channels, columns='cid,priv,act,ub,cap,lb,rb,bf,'
                              'fr,annotation,alias',
            sort_dict=sort_dict)

    def print_channels_unbalanced(self, unbalancedness, sort_string='rev_ub'):
        """
        Prints unbalanced channels with
        |unbalancedness(channel)| > unbalancedness.

        :param unbalancedness: float
        :param sort_string: str
        """

        channels = self._add_channel_annotations(
            self.node.get_unbalanced_channels(unbalancedness))

        sort_string, reverse_sorting = self._sorting_order(sort_string)
        sort_dict = {
            'function': lambda x: x[1][sort_string],
            'string': sort_string,
            'reverse': reverse_sorting,
        }

        self._print_channels(
            channels, columns='cid,ub,cap,lb,rb,bf,fr,annotation,alias',
            sort_dict=sort_dict)

    def print_channels_inactive(self, sort_string='lup'):
        """
        Prints all inactive channels.

        :param sort_string: str
        """

        channels = self._add_channel_annotations(
            self.node.get_inactive_channels())

        sort_string, reverse_sorting = self._sorting_order(sort_string)
        sort_dict = {
            'function': lambda x: (-x[1]['private'], x[1][sort_string]),
            'string': sort_string,
            'reverse': reverse_sorting,
        }

        self._print_channels(
            channels, columns='cid,lup,priv,ini,age,ub,cap,lb,rb,'
                              'sr/w,annotation,alias',
            sort_dict=sort_dict)

    def print_channels_forwardings(self, time_interval_start,
                                   time_interval_end, sort_string):

        """
        Prints forwarding statistics for each channel.

        :param time_interval_start: int
        :param time_interval_end: int
        :param sort_string: str
        """

        channels = get_forwarding_statistics_channels(
            self.node, time_interval_start, time_interval_end)

        channels = self._add_channel_annotations(channels)

        sort_string, reverse_sorting = self._sorting_order(sort_string)
        sort_dict = {
            'function': lambda x: (
                float('inf') if math.isnan(x[1][sort_string])
                else x[1][sort_string],
                x[1][PRINT_CHANNELS_FORMAT['nfwd']['dict_key']],
                x[1][PRINT_CHANNELS_FORMAT['ub']['dict_key']]
                ),
            'string': sort_string,
            'reverse': reverse_sorting,
        }

        self._print_channels(
            channels,
            columns='cid,nfwd,age,fees,f/w,flow,ub,bwd,r,'
                    'cap,bf,fr,annotation,alias',
            sort_dict=sort_dict)

    def _add_channel_annotations(self, channels):
        """
        Appends metadata to existing channel dicts from the configuration file.

        :param channels: dict
        :return: dict
        """
        logger.debug("Adding annotations from file %s.", self.node.config_file)
        # mapping between the channel point and channel id
        channel_point_mapping = {k: v['channel_point'].split(':')[0]
                                 for k, v in channels.items()}
        # only read annotations if config file is given
        if self.node.config_file:
            config = settings.read_config(self.node.config_file)
            annotations = config['annotations']
        else:
            annotations = {}
        channel_annotations_funding_id = {}
        channel_annotations_channel_id = {}

        for id, annotation in annotations.items():
            if len(id) == 18 and id.isnumeric():
                # valid channel id
                channel_annotations_channel_id[int(id)] = \
                    annotation
            elif len(id) == 64 and id.isalnum():
                # valid funding transaction id
                channel_annotations_funding_id[id] = \
                    annotation
            else:
                raise ValueError(
                    'First part needs to be either a channel id or the '
                    'funding transaction id. \n'
                    'The funding transaction id can be found with '
                    '`lncli listchannels` under the channel point (the '
                    'characters before the colon).'
                )

        for channel_id, channel_values in channels.items():
            # get the annotation by channel id first
            annotation = channel_annotations_channel_id.get(channel_id, None)
            # if no channel annotation, try with funding id
            if annotation is None:
                annotation = channel_annotations_funding_id.get(
                    channel_point_mapping[channel_id], None)

            if annotation is not None:
                channels[channel_id]['annotation'] = annotation
            else:
                channels[channel_id]['annotation'] = ''

        return channels

    def _print_channels(self, channels, columns, sort_dict):
        """
        General purpose channel printing.

        :param channels: dict
        :param columns: str
        :param sort_dict: dict
        """

        if not channels:
            logger.info(">>> Did not find any channels.")

        channels = OrderedDict(sorted(
            channels.items(), key=sort_dict['function'],
            reverse=sort_dict['reverse']))

        logger.info("Sorting channels by %s.", sort_dict['string'])

        logger.info("-------- Description --------")
        columns = columns.split(',')
        for column in columns:
            logger.info(
                "%-10s %s", column,
                PRINT_CHANNELS_FORMAT[column]['description'])

        logger.info("-------- Channels --------")
        # prepare the column header
        column_header = ''
        for column in columns:
            column_label = PRINT_CHANNELS_FORMAT[column]['align']
            column_width = PRINT_CHANNELS_FORMAT[column]['width']
            column_header += f"{column:{column_label}{column_width}} "

        # print the channel data
        for channel_number, (_, channel_data) in enumerate(channels.items()):
            if not channel_number % 20:
                logger.info(column_header)
            row = self._row_string(channel_data, columns)
            logger.info(row)

    @staticmethod
    def _row_string(column_values, columns):
        """
        Constructs the formatted row string for table printing.

        :param column_values: dict
        :param columns: list of str
        :return: formatted str
        """

        string = ''
        for column in columns:
            format_string = PRINT_CHANNELS_FORMAT[column]['format']
            conversion_function = PRINT_CHANNELS_FORMAT[column].get(
                'convert', lambda x: x)
            value = column_values[PRINT_CHANNELS_FORMAT[column]['dict_key']]
            converted_value = conversion_function(value)
            string += f"{converted_value:{format_string}} "

        return string

    @staticmethod
    def _sorting_order(sort_string):
        """
        Determines the sorting string and the sorting order.

        If sort_string starts with 'rev_', the sorting order is reversed.

        :param sort_string: str
        :return: bool
        """

        reverse_sorting = True
        if sort_string[:4] == 'rev_':
            reverse_sorting = False
            sort_string = sort_string[4:]

        sort_string = PRINT_CHANNELS_FORMAT[sort_string]['dict_key']

        return sort_string, reverse_sorting



import lndmanage.grpc_compiled.rpc_pb2 as ln

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

CLTV = 40


def set_fees_by_balancedness(
        node, base_unbalanced_msat, rate_unbalanced_decimal, base_balanced_msat, rate_balanced_decimal,
        unbalancedness=0.90):
    """
    Can be used to set fees differently for balanced and unbalanced channels.
    TODO: refactor out lnd api calls into :class:`lib.node.LndNode`

    :param node: :class:`lib.node.LndNode` instance
    :param base_unbalanced_msat: int
    :param rate_unbalanced_decimal: float (e.g. 0.00001 = 0.001 %)
    :param base_balanced_msat: int
    :param rate_balanced_decimal: float (e.g. 0.00001 = 0.001 %)
    :param unbalancedness: float between 0 ... 1 (0: very balanced, 1: very unbalanced)
    """

    channels = node.get_unbalanced_channels()
    logger.info(f"-------- unbalanced channels (|ub| > {unbalancedness}) ---------")
    channels.sort(key=lambda x: abs(x['unbalancedness']), reverse=True)
    print_divide = 0

    for c in channels:
        logger.info(f"|ub|: {abs(c['unbalancedness']):1.4f} c: {c['chan_id']} a: {c['alias']}")

        cp_parts = c['channel_point'].split(':')

        channel_point = ln.ChannelPoint(funding_txid_str=cp_parts[0], output_index=int(cp_parts[1]))

        if abs(c['unbalancedness']) > unbalancedness:  # unbalanced
            new_base = base_unbalanced_msat
            new_rate = rate_unbalanced_decimal
        else:  # balanced
            if print_divide == 0:
                logger.info(f"-------- balanced channels (|ub| < {unbalancedness}) --------")
                print_divide = 1

            new_base = base_balanced_msat
            new_rate = rate_balanced_decimal

        update_request = ln.PolicyUpdateRequest(
            chan_point=channel_point,
            base_fee_msat=new_base,
            fee_rate=new_rate,
            time_lock_delta=CLTV,
        )
        logger.debug(update_request)
        # node._stub.UpdateChannelPolicy(request=update_request)


if __name__ == '__main__':
    from lndmanage.lib.node import LndNode
    import logging.config
    from lndmanage import settings

    logging.config.dictConfig(settings.logger_config)

    nd = LndNode()

    set_fees_by_balancedness(
        nd, base_unbalanced_msat=0, rate_unbalanced_decimal=0.000001,
        base_balanced_msat=40, rate_balanced_decimal=0.000050)




import logging
from collections import OrderedDict, defaultdict

import numpy as np

from lndmanage.lib.node import LndNode
from lndmanage import settings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

np.warnings.filterwarnings('ignore')

# nearest neighbor weight for flow-analysis ~1/(avg. degree)
NEIGHBOR_WEIGHT = 0.1
# nearest neighbor weight for flow-analysis ~1/(avg. degree)^2
NEXT_NEIGHBOR_WEIGHT = 0.02


def nan_to_zero(number):
    """
    Converts float('nan') to 0
    :param number: float
    :return: float
    """
    if number is np.nan or number != number:
        return 0.0
    else:
        return number


class ForwardingAnalyzer(object):
    """
    Analyzes forwardings for single channels.
    """
    def __init__(self, node):
        self.node = node
        self.forwarding_events = self.node.get_forwarding_events()
        self.channels = {}
        self.total_forwarding_amount_sat = 0
        self.total_forwarding_fees_msat = 0
        self.forwardings = 0
        self.cumulative_effective_fee = 0
        self.timestamp_first_send = 1E10  # somewhere in future
        self.timestamp_last_send = 0  # at beginning of time
        self.max_time_interval = None

    def initialize_forwarding_data(self, time_start, time_end):
        """
        Initializes the channel statistics objects with data from
        the forwardings.

        :param time_start: time interval start, unix timestamp
        :param time_end: time interval end, unix timestamp
        """
        for f in self.forwarding_events:
            if time_start < f['timestamp'] < time_end:
                # make a dictionary entry for unknown channels
                channel_id_in = f['chan_id_in']
                channel_id_out = f['chan_id_out']

                if channel_id_in not in self.channels.keys():
                    self.channels[channel_id_in] = ChannelStatistics(
                        channel_id_in)
                if channel_id_out not in self.channels.keys():
                    self.channels[channel_id_out] = ChannelStatistics(
                        channel_id_out)

                self.channels[channel_id_in].inward_forwardings.append(
                    f['amt_in'])
                self.channels[channel_id_out].outward_forwardings.append(
                    f['amt_out'])
                self.channels[channel_id_out].absolute_fees.append(
                    f['fee_msat'])
                self.channels[channel_id_out].effective_fees.append(
                    f['effective_fee'])
                self.channels[channel_id_out].timestamps.append(
                    f['timestamp'])

                self.total_forwarding_amount_sat += f['amt_in']
                self.total_forwarding_fees_msat += f['fee_msat']
                self.forwardings += 1
                self.cumulative_effective_fee += f['effective_fee']

    def get_forwarding_statistics_channels(self):
        """
        Prepares the forwarding statistics for each channel.
        :return: dict: statistics with channel_id as keys
        """
        channel_statistics = {}

        for k, c in self.channels.items():

            try:
                timestamp_first_send = min(c.timestamps)
                if self.timestamp_first_send > timestamp_first_send:
                    self.timestamp_first_send = timestamp_first_send
            except ValueError:
                pass
            try:
                timestamp_last_send = max(c.timestamps)
                if self.timestamp_last_send < timestamp_last_send:
                    self.timestamp_last_send = timestamp_last_send
            except ValueError:
                pass

            channel_statistics[k] = {
                'effective_fee': c.effective_fee(),
                'fees_total': c.fees_total(),
                'flow_direction': c.flow_direction(),
                'mean_forwarding_in': c.mean_forwarding_in(),
                'mean_forwarding_out': c.mean_forwarding_out(),
                'median_forwarding_in': c.median_forwarding_in(),
                'median_forwarding_out': c.median_forwarding_out(),
                'number_forwardings': c.number_forwardings(),
                'largest_forwarding_amount_in':
                    c.largest_forwarding_amount_in(),
                'largest_forwarding_amount_out':
                    c.largest_forwarding_amount_out(),
                'total_forwarding_in': c.total_forwarding_in(),
                'total_forwarding_out': c.total_forwarding_out(),
            }

        # determine the time interval starting with the first forwarding
        # to the last forwarding in the analyzed time interval determined
        # by time_start and time_end
        self.max_time_interval = \
            (self.timestamp_last_send - self.timestamp_first_send) \
            / (24 * 60 * 60)

        return channel_statistics

    def get_forwarding_statistics_nodes(self, sort_by='total_forwarding'):
        """
        Calculates forwarding statistics based on single channels for
        their nodes.

        :param sort_by: str, abbreviation for the dict key
        :return:
        """
        channel_statistics = self.get_forwarding_statistics_channels()
        closed_channels = self.node.get_closed_channels()
        open_channels = self.node.get_open_channels()
        logger.debug(f"Number of channels with known forwardings: "
                     f"{len(closed_channels) + len(open_channels)} "
                     f"(thereof {len(closed_channels)} closed channels).")

        node_statistics = OrderedDict()

        # go through channel statistics and calculate node statistics
        for k, n in channel_statistics.items():
            # historic node data can be outdated, so we need to take care
            # that the remote pub key is known,
            # otherwise it is useless information

            channel_data = open_channels.get(k, None)
            if not channel_data:
                channel_data = closed_channels.get(k, None)

            if channel_data:
                remote_pubkey = channel_data['remote_pubkey']
                if remote_pubkey not in node_statistics.keys():
                    node_statistics[remote_pubkey] = {
                        'total_forwarding_in': n['total_forwarding_in'],
                        'total_forwarding_out': n['total_forwarding_out'],
                        'total_forwarding':
                            n['total_forwarding_in'] +
                            n['total_forwarding_out'],
                    }
                else:
                    node_statistics[remote_pubkey]['total_forwarding_in'] \
                        += n['total_forwarding_in']
                    node_statistics[remote_pubkey]['total_forwarding_out'] \
                        += n['total_forwarding_out']
                    node_statistics[remote_pubkey]['total_forwarding'] \
                        += (n['total_forwarding_in'] +
                            n['total_forwarding_out'])

        for k, n in node_statistics.items():
            tot_in = n['total_forwarding_in']
            tot_out = n['total_forwarding_out']
            node_statistics[k]['flow_direction'] = \
                -((float(tot_in) / (tot_in + tot_out)) - 0.5) / 0.5

        sorted_dict = OrderedDict(
            sorted(node_statistics.items(), key=lambda x: -x[1][sort_by]))

        return sorted_dict

    def simple_flow_analysis(self, last_forwardings_to_analyze=100):
        """
        Takes each forwarding event and determines the set of incoming nodes
        (up to second nearest neighbors) and outgoing nodes
        (up to second nearest neighbors) and does a frequency analyis of both
        sets, assigning a probability that a certain node was involved in
        the forwarding process.

        These probability sets are then used to take the difference between
        both, excluding effectively the large hubs. The sets can be weighted
        by a quantity of the forwarding process, e.g. the forwarding fee.

        Two lists are returned, the list of demand for sending and a list of
        nodes with demand of receiving.
        :param last_forwardings_to_analyze: int, number of last forwardings,
                                                 which should be analyzed

        :return: sending list, receiving list
        """
        # TODO: refine with channel capacities
        # TODO: refine with update time
        # TODO: refine with route calculation
        # TODO: fine tune NEIGHBOR_WEIGHT, NEXT_NEIGHBOR_WEIGHT

        logger.info("Doing simple forwarding analysis.")

        total_incoming_neighbors = defaultdict(float)
        total_outgoing_neighbors = defaultdict(float)

        logger.info(f"Total forwarding events found: "
                    f"{len(self.forwarding_events)}.")
        logger.info(f"Carrying out flow analysis for last "
                    f"{last_forwardings_to_analyze} forwarding events.")
        number_progress_report = last_forwardings_to_analyze // 10

        for nf, f in enumerate(
                self.forwarding_events[-last_forwardings_to_analyze:]):

            # report progress
            if nf % number_progress_report == 0:
                logger.info(
                    f"Analysis progress: "
                    f"{100 * float(nf) / last_forwardings_to_analyze}%")

            chan_id_in = f['chan_id_in']
            chan_id_out = f['chan_id_out']

            edge_data_in = self.node.network.edges.get(chan_id_in, None)
            edge_data_out = self.node.network.edges.get(chan_id_out, None)

            if edge_data_in is not None and edge_data_out is not None:
                # determine incoming node
                incoming_node_pub_key = edge_data_in['node1_pub'] \
                    if edge_data_in['node1_pub'] != self.node.pub_key \
                    else edge_data_in['node2_pub']

                outgoing_node_pub_key = edge_data_out['node1_pub'] \
                    if edge_data_out['node1_pub'] != self.node.pub_key \
                    else edge_data_out['node2_pub']

                # nodes involved in the forwarding process should be removed
                excluded_nodes = [self.node.pub_key, incoming_node_pub_key,
                    outgoing_node_pub_key]

                # determine all the nearest and second nearest
                # neighbors (they may appear more than once)
                incoming_neighbors = self.__determine_joined_neighbors(
                    incoming_node_pub_key, excluded_nodes=excluded_nodes)
                outgoing_neighbors = self.__determine_joined_neighbors(
                    outgoing_node_pub_key, excluded_nodes=excluded_nodes)

                # do a symmetric difference of node sets with weights
                symmetric_difference_weights = self.__symmetric_difference(
                    incoming_neighbors, outgoing_neighbors)

                final_outgoing_nodes = self.__filter_nodes(
                    symmetric_difference_weights, return_positive_weights=True)
                final_incoming_nodes = self.__filter_nodes(
                    symmetric_difference_weights,
                    return_positive_weights=False)

                # normalize the weights
                normalized_incoming = self.__normalize_neighbors(
                    final_incoming_nodes)
                normalized_outgoing = self.__normalize_neighbors(
                    final_outgoing_nodes)

                # set weight for each forwarding event
                weight = 1
                # alternatively:
                # weight = f['amt_in']
                # weight = f['fee_msat']

                for n, nv in normalized_incoming.items():
                    total_incoming_neighbors[n] += nv * weight

                for n, nv in normalized_outgoing.items():
                    total_outgoing_neighbors[n] += nv * weight

        # sort according to weights
        total_incoming_node_dict = self.__weighted_neighbors_to_sorted_dict(
            total_incoming_neighbors)
        total_outgoing_node_dict = self.__weighted_neighbors_to_sorted_dict(
            total_outgoing_neighbors)

        return total_incoming_node_dict, total_outgoing_node_dict

    @staticmethod
    def __weighted_neighbors_to_sorted_dict(node_dict):
        """
        Converts a node weight dictionary to a sorted list of
        nodes with weights.

        :param node_dict: dict with key-value pairs of node_pub_key and weight
        :return: sorted list
        """
        sorted_nodes_dict = OrderedDict()
        node_list = [(n, nv) for n, nv in node_dict.items()]
        node_list_sorted = sorted(node_list, key=lambda x: x[1], reverse=True)
        for n, nv in node_list_sorted:
            sorted_nodes_dict[n] = {'weight': nv}
        return sorted_nodes_dict

    def __determine_joined_neighbors(self, node_pub_key, excluded_nodes):
        """
        Determines the joined set of nearest and second neighbors and assigns
        a weight to every node dependent how often they appear.

        :param node_pub_key: str, public key of the home node
        :param excluded_nodes: list of str, public keys of excluded nodes
                                            in analysis
        :return: dict, keys: node_pub_keys, values: weights
        """

        neighbors = list(self.node.network.neighbors(node_pub_key))
        second_neighbors = list(
            self.node.network.second_neighbors(node_pub_key))

        # determine node_weights
        neighbor_weights = self.__analyze_neighbors(
            neighbors, excluded_nodes=excluded_nodes, weight=NEIGHBOR_WEIGHT)

        second_neighbor_weights = self.__analyze_neighbors(
            second_neighbors, excluded_nodes=excluded_nodes,
            weight=NEXT_NEIGHBOR_WEIGHT)

        # print("number of neighboring nodes", len(neighbor_weights.keys()))
        # print("number of second neighbor nodes",
        #       len(second_neighbor_weights.keys()))

        joined_neighbors = self.__join_neighbors(
            neighbor_weights, second_neighbor_weights)
        # print("number of joined neighbors", len(joined_neighbors.keys()))
        return joined_neighbors

    @staticmethod
    def __normalize_neighbors(neighbors_weight_dict):
        """
        Normalizes the weights of the neighbors to the total weight
        of all neighbors.

        :param neighbors_weight_dict: dict, keys: node_pub_keys,
                                            values: weights
        :return: dict, keys: node_pub_keys, values: normalized weights
        """
        total_weight = 0
        normalized_dict = {}
        # determine total weight
        for n, nv in neighbors_weight_dict.items():
            total_weight += nv
        # normalize
        for n, nv in neighbors_weight_dict.items():
            normalized_dict[n] = float(nv) / total_weight

        return normalized_dict

    @staticmethod
    def __analyze_neighbors(neighbors, excluded_nodes, weight):
        """
        Analyzes a node dict for the frequency of nodes and gives them
        a weight. An upper bound of the weights is set.

        :param neighbors: list of node_pub_keys
        :param excluded_nodes: excluded node_pub_keys for analysis
        :param weight: float, weight for each individual appearance of a node
        :return: dict, keys: node_pub_keys, values: node weights
        """
        node_weights = {}

        for n in neighbors:
            if n not in excluded_nodes:
                if n in node_weights.keys():
                    node_weights[n] = min(node_weights[n] + weight, 1.0)
                else:
                    node_weights[n] = weight

        return node_weights

    @staticmethod
    def __join_neighbors(first_neighbor_dict, second_neighbor_dict):
        """
        Joins two node weight dicts together.
        :param first_neighbor_dict: dict, keys: node_pub_keys,
                                          values: node weights
        :param second_neighbor_dict: dict, keys: node_pub_keys,
                                           values: node weights
        :return: dict, keys: node_pub_keys, values: node weights
        """
        # make a copy of the first node weight dict
        joined_neighbor_dict = dict(first_neighbor_dict)

        # add all the nodes from the second dict
        for n, v in second_neighbor_dict.items():
            if n in joined_neighbor_dict:
                joined_neighbor_dict[n] = max(joined_neighbor_dict[n],
                                              second_neighbor_dict[n])
            else:
                joined_neighbor_dict[n] = second_neighbor_dict[n]

        return joined_neighbor_dict

    @staticmethod
    def __symmetric_difference(first_neighbors_dict, second_neighbors_dict):
        """
        Calculates the difference of weights of first and second node dicts,
        doing also a symmetric difference between the sets of nodes.

        :param first_neighbors_dict: dict, keys: node_pub_keys,
                                           values: node weights
        :param second_neighbors_dict: dict, keys: node_pub_keys,
                                            values: node weights
        :return: dict, keys: node_pub_keys, values: node weights
        """
        first_nodes = set(first_neighbors_dict.keys())
        second_nodes = set(second_neighbors_dict.keys())
        # nodes_union = first_nodes.union(second_nodes)
        # print("unique nodes", len(nodes_union))

        nodes_intersection = first_nodes.intersection(second_nodes)
        # print("common nodes", len(nodes_intersection))

        unique_first_nodes = first_nodes - nodes_intersection
        unique_second_nodes = second_nodes - nodes_intersection

        diff_neighbor_dict = {}
        for n in unique_first_nodes:
            diff_neighbor_dict[n] = first_neighbors_dict[n]

        for n in unique_second_nodes:
            diff_neighbor_dict[n] = second_neighbors_dict[n]

        for n in nodes_intersection:
            delta = first_neighbors_dict[n] - second_neighbors_dict[n]
            if delta != 0.0:
                diff_neighbor_dict[n] = delta

        return diff_neighbor_dict

    @staticmethod
    def __filter_nodes(node_weights, return_positive_weights=True):
        """
        Filters out nodes with positive or negative weights and
        takes the absolute.

        :param node_weights: dict
        :param return_positive_weights: bool, if True returns nodes with
                                              positive weights,
                                              if False negative weights
        :return:
        """
        new_node_weights = {}
        for n, nv in node_weights.items():
            if nv > 0 and return_positive_weights:
                new_node_weights[n] = nv
            elif nv < 0 and not return_positive_weights:
                new_node_weights[n] = -nv
        return new_node_weights


class ChannelStatistics(object):
    """
    Functionality to analyze the forwardings of a single channel.
    """
    def __init__(self, channel_id):
        self.channel_id = channel_id

        self.inward_forwardings = []
        self.outward_forwardings = []

        self.timestamps = []
        self.absolute_fees = []
        self.effective_fees = []

    def total_forwarding_in(self):
        return sum(self.inward_forwardings)

    def total_forwarding_out(self):
        return sum(self.outward_forwardings)

    def mean_forwarding_in(self):
        return np.mean(self.inward_forwardings)

    def mean_forwarding_out(self):
        return np.mean(self.outward_forwardings)

    def median_forwarding_in(self):
        return np.median(self.inward_forwardings)

    def median_forwarding_out(self):
        return np.median(self.outward_forwardings)

    def fees_total(self):
        return sum(self.absolute_fees)

    def effective_fee(self):
        return np.mean(self.effective_fees)

    def largest_forwarding_amount_out(self):
        return max(self.outward_forwardings, default=float('nan'))

    def largest_forwarding_amount_in(self):
        return max(self.inward_forwardings, default=float('nan'))

    def flow_direction(self):
        total_in = self.total_forwarding_in()
        total_out = self.total_forwarding_out()
        return -((float(total_in) / (total_in + total_out)) - 0.5) / 0.5

    def number_forwardings(self):
        return len(self.inward_forwardings) + len(self.outward_forwardings)


def get_forwarding_statistics_channels(node, time_interval_start,
                                       time_interval_end):
    """
    Joins data from listchannels and fwdinghistory to have a extended
    information about a channel.

    :param node: :class:`lib.node.Node`
    :param time_interval_start: unix timestamp
    :param time_interval_end: unix timestamp
    :return: dict of channel information with channel_id as keys
    """
    forwarding_analyzer = ForwardingAnalyzer(node)
    forwarding_analyzer.initialize_forwarding_data(
        time_interval_start, time_interval_end)

    # dict with channel_id keys
    statistics = forwarding_analyzer.get_forwarding_statistics_channels()
    logger.debug(f"Time interval (between first and last forwarding) is "
                 f"{forwarding_analyzer.max_time_interval:6.2f} days.")
    # join the two data sets:
    channels = node.get_unbalanced_channels(unbalancedness_greater_than=0.0)

    # TODO: improve this code, don't repeat
    for k, c in channels.items():
        try:  # channel forwarding statistics exists
            chan_stats = statistics[c['chan_id']]
            c['bandwidth_demand'] = max(
                nan_to_zero(chan_stats['mean_forwarding_in']),
                nan_to_zero(chan_stats['mean_forwarding_out'])
            ) / c['capacity']
            c['fees_total'] = chan_stats['fees_total']
            # time interval may be zero, to avoid zero division, replace by NaN
            try:
                c['fees_total_per_week'] = chan_stats['fees_total'] \
                    / (forwarding_analyzer.max_time_interval / 7)
            except ZeroDivisionError:
                c['fees_total_per_week'] = float('nan')
            c['flow_direction'] = chan_stats['flow_direction']
            c['median_forwarding_in'] = chan_stats['median_forwarding_in']
            c['median_forwarding_out'] = chan_stats['median_forwarding_out']
            c['mean_forwarding_in'] = chan_stats['mean_forwarding_in']
            c['mean_forwarding_out'] = chan_stats['mean_forwarding_out']
            c['number_forwardings'] = chan_stats['number_forwardings']
            c['largest_forwarding_amount_in'] = \
                chan_stats['largest_forwarding_amount_in']
            c['largest_forwarding_amount_out'] = \
                chan_stats['largest_forwarding_amount_out']
            c['total_forwarding_in'] = chan_stats['total_forwarding_in']
            c['total_forwarding_out'] = chan_stats['total_forwarding_out']

            # action required if flow same direction as unbalancedness
            # or bandwidth demand too high

            # TODO: refine 'action_required' by better metric
            if c['unbalancedness'] * c['flow_direction'] > 0 and abs(
                    c['unbalancedness']) > settings.UNBALANCED_CHANNEL:
                c['action_required'] = True
            else:
                c['action_required'] = False
            if c['bandwidth_demand'] > 0.5:
                c['action_required'] = True

        except KeyError:  # no forwarding statistics on channel is available
            c['bandwidth_demand'] = 0
            c['fees_total'] = 0
            c['fees_total_per_week'] = 0
            c['flow_direction'] = float('nan')
            c['median_forwarding_out'] = float('nan')
            c['median_forwarding_in'] = float('nan')
            c['mean_forwarding_out'] = float('nan')
            c['mean_forwarding_in'] = float('nan')
            c['number_forwardings'] = 0
            c['largest_forwarding_amount_in'] = float('nan')
            c['largest_forwarding_amount_out'] = float('nan')
            c['total_forwarding_in'] = float('nan')
            c['total_forwarding_out'] = float('nan')
            if abs(c['unbalancedness']) > settings.UNBALANCED_CHANNEL:
                c['action_required'] = True
            else:
                c['action_required'] = False
            if c['bandwidth_demand'] > 0.5:
                c['action_required'] = True
    return channels


if __name__ == '__main__':
    import time
    import logging.config
    logging.config.dictConfig(settings.logger_config)
    logger = logging.getLogger()

    nd = LndNode()
    fa = ForwardingAnalyzer(nd)
    fa.initialize_forwarding_data(time_start=0, time_end=time.time())
    print(fa.simple_flow_analysis())


import collections


def convert_dictionary_number_strings_to_ints(data):
    """
    Converts the values of str type to int type where possible.

    :param data: dict
    :return: converted dict
    """
    if isinstance(data, str):
        try:
            data = int(data)
        except ValueError:
            data = str(data)
        return data
    elif isinstance(data, collections.Mapping):
        return dict(map(convert_dictionary_number_strings_to_ints, data.items()))
    elif isinstance(data, collections.Iterable):
        return type(data)(map(convert_dictionary_number_strings_to_ints, data))
    else:
        return data
